searchData={"items":[{"type":"module","title":"Hardhat","doc":"An opinionated, production-ready HTTP client for Elixir services.","ref":"Hardhat.html"},{"type":"module","title":"What's included - Hardhat","doc":"- [X] Connection pooling per-client module\n- [X] Integration with `telemetry` and `opentelemetry` instrumentation\n- [X] Circuit breaking for repeatedly failed requests\n- [X] Automatic retries for failed requests\n- [X] Timeout and `deadline` support","ref":"Hardhat.html#module-what-s-included"},{"type":"module","title":"Why Hardhat? - Hardhat","doc":"TODO","ref":"Hardhat.html#module-why-hardhat"},{"type":"module","title":"Installation - Hardhat","doc":"Add `hardhat` to the dependencies in your `mix.exs`:\n\n```elixir\n  def deps do\n    [\n      {:hardhat, \"~> 1.0.0\"}\n    ]\n  end\n```","ref":"Hardhat.html#module-installation"},{"type":"module","title":"Getting started - Hardhat","doc":"`Hardhat` is designed to be easy for creating quick wrappers around HTTP APIs,\nbut includes many options for customization. To define a simple client, do something like the following:\n\n```elixir\n# Define a client module:\ndefmodule SocialMediaAPI do\n  use Hardhat\nend\n\n# Add it to your supervisor (required):\ndefmodule MyApp.Sup do\n  use Supervisor\n\n  def start_link(init_arg) do\n    Supervisor.start_link(__MODULE__, init_arg, name: __MODULE__)\n  end\n\n  @impl true\n  def init(_init_arg) do\n    children = [\n      SocialMediaAPI\n    ]\n\n    Supervisor.init(children, strategy: :one_for_one)\n  end\nend\n\n# Use your client to make requests:\nSocialMediaAPI.get(\"http://media-api.social/posts\")\n```\n\nAs mentioned in the example above, it is **imperative** for you to supervise the client module that includes the `use Hardhat` macro. Without starting the client under supervision, you will not be able to make requests. See [Connection pools](#module-connection-pools) below for more information.","ref":"Hardhat.html#module-getting-started"},{"type":"module","title":"General behavior - Hardhat","doc":"`Hardhat` is built on top of `Tesla`, and uses `Finch` as the adapter. Because\n`Tesla` is the foundation, you are welcome to use publicly available\n`Tesla.Middleware` modules in your `Hardhat`-based client (with the exception\nthat [we recommend](#module-timeouts-and-deadlines) you use\n`Hardhat.Middleware.Timeout` instead of `Tesla.Middleware.Timeout`).\n\n```elixir\ndefmodule SomeJSONAPI do\n  use Hardhat\n\n  plug Tesla.Middleware.BaseUrl, \"https://my-json.api/\"\n  plug Tesla.Middleware.JSON\nend\n```\n\nIn addition to the adapter selection and default `Tesla` behavior,\n`use Hardhat` will inject the common functionality [listed above](#module-what-s-included) *after* any middleware that you supply via [`plug`](`Tesla.Builder.plug/2`). The current list is as follows:\n\n* `Hardhat.Middleware.DeadlinePropagation`\n* `Tesla.Middleware.Retry` ([see below](#module-retries))\n* Either `Tesla.Middleware.Fuse` or `Hardhat.Middleware.Regulator` ([See below](#module-failure-detection)\n* `Tesla.Middleware.Telemetry` ([see below](#module-telemetry-and-tracing))\n* `Tesla.Middleware.OpenTelemetry` ([see below](#module-telemetry-and-tracing))\n* `Hardhat.Middleware.PathParams`","ref":"Hardhat.html#module-general-behavior"},{"type":"module","title":"Connection pools - Hardhat","doc":"As mentioned above, `Hardhat` uses `Finch` as the adapter. By [default](`Hardhat.Defaults.pool_options/1`), `Hardhat` specifies a connection pool of size `10` but sets no [other options](`Finch.start_link/1`) on the adapter. The name of the `Finch` process is proscribed by the `use Hardhat` macro, but you can set any other options for the pool that you like, including creating more than one pool or setting the HTTP protocol or TLS options by overriding the `pool_options/1` function.\n\n```elixir\ndefmodule H2Client do\n  use Hardhat\n\n  # This function overrides the default coming from `Hardhat.Defaults`.\n  # The `overrides` will be passed from your process supervision initial\n  # arguments.\n  def pool_options(_overrides \\\\ []) do\n    %{\n      # By default we'll use HTTP/2, with 3 pools of one connection each\n      :default => [\n        protocol: :http2,\n        count: 3\n      ],\n      # For this host only, we're using HTTP/1.1 and a single pool of 20\n      # connections\n      \"https://some-http1-only-host.com/\" => [\n        size: 20\n      ]\n    }\n  end\nend\n```","ref":"Hardhat.html#module-connection-pools"},{"type":"module","title":"Telemetry and tracing - Hardhat","doc":"`Hardhat` includes the stock `Tesla.Middleware.Telemetry` for injecting your own metrics and monitoring systems into its operation. The events emitted by this middleware are:\n\n* `[:tesla, :request, :start]` - at the beginning of the request\n* `[:tesla, :request, :stop]` - at the completion of the request\n* `[:tesla, :request, :exception]` - when a non-HTTP-status exception occurs\n\n```elixir\ndefmodule TelemetryClient do\n  use Hardhat\nend\n\n:telemetry.attach(\n  \"my handler\",\n  [:tesla, :request, :stop],\n  fn _event, measures, _metadata, _config ->\n     # Don't do this, attach to your metrics system instead\n     IO.puts(\"Made a request in #{measures.duration}\")\n  end,\n  nil\n)\n```\n\n`Hardhat` wraps each request in an [OpenTelemetry](`Tesla.Middleware.OpenTelemetry`) span and propagates the trace context to the destination host. It does not currently expose the ability to change the span name in the trace, but it will observe any [path parameters](`Hardhat.Middleware.PathParams`) you interpolate into the URL so that similar spans can be easily aggregated.","ref":"Hardhat.html#module-telemetry-and-tracing"},{"type":"module","title":"Failure detection - Hardhat","doc":"`Hardhat` provides two different failure detection and backoff strategies:\n\n* Static circuit breaking with `:fuse` (`Tesla.Middleware.Fuse`)\n* Dynamic request rate regulation ([AIMD](`Regulator.Limit.AIMD`) with `Regulator` (`Hardhat.Middleware.Regulator`)\n\nThese strategies cannot be used together safely, so you must choose one when defining your client. If your needs are simple and hard failures are relatively rare, `:fuse` is an easier strategy to comprehend and implement because it uses a straightforward failure-counting algorithm, and completely turns off requests when the configured threshold is reached. If you have a more complicated setup or high traffic, and do not want to spend as much time tuning your failure behavior, the `:regulator` strategy might be for you. `Regulator` allows your client to adapt to rapidly changing conditions by reducing the amount of concurrent work in the presence of failure, without causing a hard stop to activity. On the other hand, if your concurrent utilization is low, it might also bound your maximum concurrency even when requests are not failing.\n\n#","ref":"Hardhat.html#module-failure-detection"},{"type":"module","title":"The Fuse strategy - Hardhat","doc":"The `:fuse` failure detection strategy is configured with two functions in your client which have default implementations that are injected at compile-time:\n\n* [`fuse_opts()`](`Hardhat.Defaults.fuse_opts/1`) - configuration for the middleware\n* [`should_melt(result)`](`Hardhat.Defaults.should_melt/1`) - whether the result of the request is considered a failure\n\nYou can override their default behavior by redefining the functions:\n\n```elixir\n# This module uses `:fuse` for failure detection and backoff.\ndefmodule HardCutoffClient do\n  use Hardhat # defaults to `:fuse`\n\n  # This is also valid:\n  # use Hardhat, strategy: :fuse\n\n  # Customize fuse's circuit-breaking behavior\n  def fuse_opts do\n    [\n      # 10 failed requests in 0.5sec flips the breaker, which resets after 1sec\n      opts: {{:standard, 10, 500}, {:reset, 1_000}},\n      # Return the error that caused the fuse to break\n      keep_original_error: true,\n      # Use our custom failure test\n      should_melt: &__MODULE__.should_melt/1\n      # Go fast and a little loose\n      mode: :async_dirty\n    ]\n  end\n\n  # Customize how responses are determined to be failures,\n  # in this case only TCP/adapter-type errors are considered\n  # failures, any valid response is fine.\n  def should_melt(result) do\n    case result do\n      {:error, _} -> true\n      {:ok, %Tesla.Env{}} -> false\n    end\n  end\nend\n```\n\n#","ref":"Hardhat.html#module-the-fuse-strategy"},{"type":"module","title":"The Regulator strategy - Hardhat","doc":"The `:regulator` failure detection strategy is configured with two functions in your client which have default implementations that are injected at compile-time:\n\n* [`regulator_opts()`](`Hardhat.Defaults.regulator_opts/1`) - configuration for the middleware\n* [`should_regulate(result)`](`Hardhat.Defaults.should_regulate/1`) - whether the result of the request is considered a failure\n\nYou can override their default behavior by redefining the functions:\n\n```elixir\n# This module uses `Regulator` for failure detection and backoff\ndefmodule DynamicRegulationClient do\n  use Hardhat, strategy: :regulator # overrides default of `:fuse`\n\n  # Customize Regulator's ramp-up and backoff strategy\n  def regulator_opts do\n    [\n      # Backoff on failure by half instead of 90%\n      backoff_ratio: 0.5,\n      should_regulate: &__MODULE__.should_regulate/1\n    ]\n  end\n\n  # Customize how responses are determined to be failures,\n  # in this case TCP/adapter-level errors are considered failures,\n  # as well as HTTP `429 Too Many Requests` responses.\n  def should_regulate(result) do\n    case result do\n      {:error, _} -> true\n      {:ok, %Tesla.Env{status: 429}} -> true\n      {:ok, %Tesla.Env{}} -> false\n    end\n  end\nend\n```","ref":"Hardhat.html#module-the-regulator-strategy"},{"type":"module","title":"Retries - Hardhat","doc":"TODO: default options, note about interactions with failure detection\n\n`Hardhat` injects automatic retries on your requests using `Tesla.Middleware.Retry`. Retries are configured with two functions in your client which have default implementations that are injected at compile-time:\n\n* [`retry_opts()`](`Hardhat.Defaults.retry_opts/1`) - configuration for the middleware\n* [`should_retry(result)`](`Hardhat.Defaults.should_retry/1`) - whether the result of the request can be retried\n\nYou can override their default behavior by redefining the functions:\n\n```elixir\n# This client retries requests\ndefmodule SomeRetriesClient do\n  use Hardhat\n\n  def retry_opts do\n    [\n      # Retry up to 5 times\n      max_retries: 5,\n      # Delay at least 75ms between attempts\n      delay: 75,\n      # Delay at most 500ms between any attempts\n      max_delay: 500,\n      # Use the default retry rubric\n      should_retry: &__MODULE__.should_retry/1,\n      # Add jitter of up to 20% to delays\n      jitter_factor: 0.2\n    ]\n  end\nend\n\n# This client disables retries entirely!\ndefmodule NoRetriesClient do\n  use Hardhat\n\n  # Override should_retry to disable retries\n  def should_retry(_), do: false\nend\n```\n\n> #","ref":"Hardhat.html#module-retries"},{"type":"module","title":"Interaction with failure detection {: .warning} - Hardhat","doc":"> Retries can interact very negatively with [failure detection](#module-failure-detection), potentially triggering backoff behavior too quickly. Be sure to avoid retrying when the failure detector returns `{:error, :unavailable}`, which indicates that the circuit breaker has blown in the `:fuse` strategy, or the limiter is out of capacity in the `:regulator` strategy.\n>\n> The default implementation of `should_retry/1` implements this behavior.","ref":"Hardhat.html#module-interaction-with-failure-detection-warning"},{"type":"module","title":"Timeouts and deadlines - Hardhat","doc":"TODO: Custom timeout middleware, `Deadline` support","ref":"Hardhat.html#module-timeouts-and-deadlines"},{"type":"module","title":"Testing - Hardhat","doc":"TODO: overriding the adapter to use a mock or bypass","ref":"Hardhat.html#module-testing"},{"type":"module","title":"Hardhat.Defaults","doc":"Contains default implementations of functions that can be overridden in\nclients that `use Hardhat`.","ref":"Hardhat.Defaults.html"},{"type":"function","title":"Hardhat.Defaults.deadline_propagation_opts/0","doc":"Default implementation of the options to `Hardhat.Middleware.DeadlinePropagation` that\nis included in the default middleware stack.\n\nThese defaults include:\n- `:header` - `\"deadline\"` is the name of the header added to the request","ref":"Hardhat.Defaults.html#deadline_propagation_opts/0"},{"type":"function","title":"Hardhat.Defaults.fuse_opts/1","doc":"Default implementation of the options to the `Tesla.Middleware.Fuse` that is\nincluded in the default middleware stack.\n\nTakes the client module name as an argument.\n\nThese defaults include:\n- `:opts` - The fuse will blow after 50 errors with 1 second, and reset after 2 seconds\n- `:keep_original_error` - The original error will be returned when a fuse first blows\n- `:should_melt` - The `should_melt/1` function defined in the client module is used\n   (by default this is `Hardhat.Defaults.should_melt/1`)\n- `:mode` - The fuse uses `:async_dirty` mode to check the failure rate, which may result\n  in some delay in blowing the fuse under high concurrency, but it will not serialize\n  checks to the fuse state through a single process\n\nSee `Tesla.Middleware.Fuse` for more information on how to configure the middleware.","ref":"Hardhat.Defaults.html#fuse_opts/1"},{"type":"function","title":"Hardhat.Defaults.pool_options/1","doc":"The default options for the connection pool that will be created when your\nclient is added to the supervision tree.\n\nThis creates a connection pool of size `10`.","ref":"Hardhat.Defaults.html#pool_options/1"},{"type":"function","title":"Hardhat.Defaults.regulator_opts/1","doc":"Default options for the `Regulator` middleware, which can be used as an\nalternative circuit-breaking strategy to `:fuse`.\n\nThe defaults include:\n* `:min_limit` - The minimum concurrency limit (defaults to 5)\n* `:initial_limit` - The initial concurrency limit when the regulator is installed (deafults to 20)\n* `:max_limit` - The maximum concurrency limit (defaults to 200)\n* `:step_increase` - The number of tokens to add when regulator is increasing the concurrency limit (defaults to 10).\n* `:backoff_ratio` - Floating point value for how quickly to reduce the concurrency limit (defaults to 0.9)\n* `:target_avg_latency` - This is the average latency in milliseconds for the system regulator is protecting. If the average latency drifts above this value Regulator considers it an error and backs off. Defaults to 5.\n* `:should_regulate` - Whether to consider the result of the request as failed, defaults to `should_regulate/1`.","ref":"Hardhat.Defaults.html#regulator_opts/1"},{"type":"function","title":"Hardhat.Defaults.retry_opts/1","doc":"Default implementation of the options to `Tesla.Middleware.Retry` that is\nincluded in the default middleware stack.\n\nThese defaults include:\n- `:delay` - The base delay in milliseconds (positive integer, defaults to 50)\n- `:max_retries` - maximum number of retries (non-negative integer, defaults to 3)\n- `:max_delay` - maximum delay in milliseconds (positive integer, defaults to 300)\n- `:should_retry` - function to determine if request should be retried, defaults to `should_retry/1`\n- `:jitter_factor` - additive noise proportionality constant (float between 0 and 1, defaults to 0.2)","ref":"Hardhat.Defaults.html#retry_opts/1"},{"type":"function","title":"Hardhat.Defaults.should_melt/1","doc":"Default implementation of the \"melt test\" for circuit breaking. This\nfunction will cause the circuit breaker to record an error when the\nresult of a request is:\n\n* A TCP-level error, e.g. `{:error, :econnrefused}`\n* An HTTP status that indicates a server error or proxy-level error (>= 500)\n* An `429 Too Many Requests` HTTP status","ref":"Hardhat.Defaults.html#should_melt/1"},{"type":"function","title":"Hardhat.Defaults.should_regulate/1","doc":"Default implementation of the \"failure test\" for dynamic regulation. This\nfunction will cause the `Regulator` to record an error when the\nresult of a request is:\n\n* A TCP-level error, e.g. `{:error, :econnrefused}`\n* An HTTP status that indicates a server error or proxy-level error (>= 500)\n* An `429 Too Many Requests` HTTP status","ref":"Hardhat.Defaults.html#should_regulate/1"},{"type":"function","title":"Hardhat.Defaults.should_retry/1","doc":"Default implementation of the \"retry test\" for retries. This\nfunction will cause requests to be retried when the result of the\nrequest is:\n\n* A TCP-level error, e.g. `{:error, :econnrefused}`\n* An HTTP status that indicates a server error or proxy-level error (>= 500)\n* An `429 Too Many Requests` HTTP status\n\nIn the case where the circuit breaker has been triggered, requests will\nnot be retried.","ref":"Hardhat.Defaults.html#should_retry/1"},{"type":"module","title":"Hardhat.Middleware.DeadlinePropagation","doc":"Propagates `Deadline` information to the server being called via\nan HTTP header. _This middleware is part of the default stack_.\n\nExpects a `:header` option, which defaults to `\"deadline\"`. See `Hardhat.Defaults.deadline_propagation_opts/0`.","ref":"Hardhat.Middleware.DeadlinePropagation.html"},{"type":"module","title":"Hardhat.Middleware.PathParams","doc":"Use templated URLs with separate params. Unlike `Tesla.Middleware.PathParams`,\nwe ensure that all parameters are URL-safe. _This middleware is part of the\ndefault stack_.","ref":"Hardhat.Middleware.PathParams.html"},{"type":"module","title":"Hardhat.Middleware.Regulator","doc":"Limits concurrent requests to the server based on a [loss based dynamic limit algorithm](`Regulator.Limit.AIMD`).\nAdditively increases the concurrency limit when there are no errors and multiplicatively\ndecrements the limit when there are errors.\n\n_This middleware is part of the default stack when giving the `strategy: :regulator` option to `use Hardhat`._\n\nSee also:\n- Configuration options: `Hardhat.Defaults.regulator_opts/1`\n- Determining what amounts to an error: `Hardhat.Defaults.should_regulate/1`","ref":"Hardhat.Middleware.Regulator.html"},{"type":"module","title":"Hardhat.Middleware.Timeout","doc":"Abort HTTP requests after the given timeout or the current `Deadline`,\nand emit `OpenTelemetry` trace events on timeouts.\n\nOptions:\n- `:timeout` - (required) timeout in milliseconds","ref":"Hardhat.Middleware.Timeout.html"},{"type":"module","title":"OpenTelemetry - Hardhat.Middleware.Timeout","doc":"Implementing a timeout necessitates moving the request into a new process,\nand then waiting on that new process's completion (or aborting after the timeout).\nSince `OpenTelemetry` tracing context is stored in the process dictionary, that\ncontext must be explicitly propagated to the new process. This middleware uses\n`OpentelemetryProcessPropagator` for this purpose.\n\nIn the event of a timeout result in this middleware, a new `timeout_exceeded` event\nwill be added to the trace. The event will include these attributes:\n\n- `module` - the client module that includes this middleware\n- `timeout` - the duration that was exceeded","ref":"Hardhat.Middleware.Timeout.html#module-opentelemetry"},{"type":"module","title":"Deadline - Hardhat.Middleware.Timeout","doc":"When the caller has set a `Deadline` for the current process, that limit will be\nrespected by this middleware. The effective timeout chosen will be the **lesser** of the time\nremaining on the current deadline and the duration given in the `:timeout` option.","ref":"Hardhat.Middleware.Timeout.html#module-deadline"}],"content_type":"text/markdown"}